/* Distributed under the GPL license */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <dirent.h>
#include <sys/stat.h>
#include <unistd.h>
#include <pthread.h>
#include <sys/time.h>

#define NAME_LENGTH 256

/* Requires only the first service found */
#define PROREQ_NORMAL 0	
/* "Requires: *" => service launched after every other */
#define REQUIRES_ALL 1	
/* Requires all service found providing the requires */
#define REQUIRES_EVERY_PROV 2 


#define MODE_START 'S'
#define MODE_KILL 'K'

#define malloc_or_die(ptr,size) do { ptr = malloc(size); if(ptr == NULL) { printf("Out of memory, Line: %d\n",__LINE__); exit(1); } } while(0)

/* ... I think I have an architectural problem.
 * three structure ( which mean three different list
 * type ) for this work seem a bit overkill */

/* Here is a simplified internal service representations :

struct service
-------------   ,-> -------------   ,->-------------
- service 1 -  /    - service 2 -  /   - service N -
-    next   ->~     -    next   ->~    -   next    -> NULL
- provides  ->+     - provides  -      - provides  -
- requires  ->|--+  - requires  -      - requires  -
------------- |  |  -------------      -------------
              |  |
struct proreq |   \
--------------+    \  struct proreq       struct pserv
- Provide 1   -     ->--------------   ,->-------------   ,->-------------   ,->-------------
- next        ->+     - Requires 1 -  /   - service 1 -  /   - service 2 -  /   - service N -
--------------- |     -   servl    ->-    -   next    ->-    -   next    ->-    -   next    ->NULL
                /     -   next     ->+    -------------      -------------      -------------
               /      -------------- |
--------------+                      /
- Provide 2   -                     /
- next        ->+     -------------+ 
--------------- |     - Requires 2 -
                /     -   servl    -
               /      -   next     ->+
--------------+       -------------- |
- Provide N   -                      /
- next        ->NULL                /
---------------       -------------+
                      - Requires N -
                      -   servl    -
                      -   next     ->NULL
                      --------------
*/   

/* List of service's pointer */
struct pserv {
	struct service * serv;  /* pointer on the service */
	struct pserv * next;    /* next service pointer in the list */
};

/* list of requires/provides */
struct proreq {
	char name[NAME_LENGTH];  /* Requires/Provide name */
	int proreq_t;            /* Requires flags */
	struct pserv * servl;    /* list of service providing the require */ 
	struct proreq * next;    /* next requires/provide in the list */
};

/* list of services */
struct service {
	char name[NAME_LENGTH];   /* name of the service */
	char file[NAME_LENGTH];   /* path to service' file */
	pthread_mutex_t mut;      /* mutex, used to protect cond */
	pthread_cond_t cond;      /* condition to tell "I'm started" change */
	int started;              /* the real "I'm started" flag */
	int old_init;             /* flag to tell "I'm not a X-Parallel-init service" */
	int count; 		  /* used for internal loop detection between service */
	struct proreq * provides; /* list of service' provides */
	struct proreq * requires; /* list of service' requires */
	struct service * next;    /* next service in the list */
};

/* head of the service list */
struct service *servlist = NULL;

/* mutex to protect console access */
pthread_mutex_t console_mutex = PTHREAD_MUTEX_INITIALIZER;

/* general current mode, MODE_START or MODE_KILL */
char mode;


/* called every time a service is started/stopped
 * update the bootsplash progressbar */
void rc_splash(char * string) {
	char * env;
	int progress;
	char buf[255];
	static pthread_mutex_t splash_mutex = PTHREAD_MUTEX_INITIALIZER;
	
	env = getenv("splash_rc");
	
	if(!env || strcmp(env,"yes")) {
		return;
	}
	
	pthread_mutex_lock(&splash_mutex);
	if((env = getenv("progress")) == NULL) 
		progress = 0;
	else
		progress = atoi(env);
	
	snprintf(buf,sizeof(buf),"%d",progress + 1);
	
	setenv("progress",buf,1);
	
	snprintf(buf,sizeof(buf),"/sbin/splash.sh %s 2>&1",string);
#ifdef DEBUG
	puts(buf);
#else
	system(buf);
#endif
	pthread_mutex_unlock(&splash_mutex);
}

/* return an sorted ( but inverted ) malloced array of file
 * found in "dir" */
char ** list_files(char * dir) {
	struct dirent **namelist;
	int n;
	int i = 0;
	char ** templist;
	n = scandir(dir, &namelist, NULL, alphasort);
	if (n < 0)
		return NULL;
	else if(n == 0) {
		free(namelist);
		return NULL;
	} else {
		malloc_or_die(templist, n * sizeof(char *) + 1);
		while(n--) {
			struct stat buf;
			char buffer[1024]; /* path are never > 255 */
			snprintf(buffer,sizeof(buffer),"%s/%s",dir,namelist[n]->d_name);
			if(stat(buffer, &buf) != 0){
				printf("Cannot stat %s\n",buffer);
				free(namelist[n]);
				continue;
			}
			if(!S_ISREG(buf.st_mode) || namelist[n]->d_name[0] != mode) {
				free(namelist[n]);
				continue;
			}
			malloc_or_die(templist[i],strlen(buffer) + 1 /* \0 */ );
			strcpy(templist[i],buffer);
			i++;
			free(namelist[n]);
		}
		/* end flag */
		templist[i] = NULL;	
		free(namelist);
		
	}
	return templist;
}

/* check if the script "path" is a new X-Parallel-init ( return 0 ),
 * or a old one ( return 1 ) */
int check_old_script(char * path) {
	FILE * file;
	char buffer[1024];

	if((file = fopen(path,"r")) == NULL) {
		printf("Cannot scan %s file: %s\n",path,strerror(errno));
		exit(1);
	}
	while(fgets(buffer,sizeof(buffer),file) != NULL) {
		if(strncmp("# X-Parallel-init",buffer,17) == 0) {
			fclose(file);
			return 0;
		}
	}
	fclose(file);
	return 1;
}

/* fill the buffer "buf" with what is on the "# Requires: " line. */
void parse_file_dep(char * path, char * buf, int buflen) {
	FILE * file;
	char buffer[1024];
	
	
	if((file = fopen(path,"r")) == NULL) {
		printf("Cannot scan %s file: %s\n",path,strerror(errno));
		exit(1);
	}
	
	while(fgets(buffer,sizeof(buffer),file) != NULL) {
		if(strncmp("# Requires:",buffer,11) == 0) {
			strncpy(buf,buffer + 11, buflen);
			fclose(file);
			return;
		}	
	}
	*buf = '\0';
	fclose(file);
}	

/* Same things, but with the "# Provides: " line */
void parse_file_provide(char * path, char * buf, int buflen) {
	FILE * file;
	char buffer[1024];


        if((file = fopen(path,"r")) == NULL) {
		printf("Cannot scan %s file: %s\n",path,strerror(errno));
                exit(1);
        }

        while(fgets(buffer,sizeof(buffer),file) != NULL) {
                if(strncmp("# Provides:",buffer,11) == 0) {
                        strncpy(buf,buffer + 11, buflen);
			fclose(file);
                        return;
                }
        }
        *buf = '\0';
	fclose(file);
}

/* fill the "list" with what is found in the string "str" 
 * Only fill the name and the flags. It does NOT resolve dependency */
void fill_proreq_list(struct proreq ** list, char * str) {
	
	if(str == NULL) {
		*list = NULL;
		return;
	}
	
	/* remove unusefull character */
	str += strspn(str," \t\n");
	
	if(!strlen(str)) {
		*list = NULL;
		return;
	}
		
	while(*str != '\0') {
		char * temp = str;
		struct proreq * prt;
		malloc_or_die(prt,sizeof(struct proreq));
		str += strcspn(str," \t\n");
		*str = '\0';
		str++;
		str += strspn(str," \t\n");
		if(*temp == '%') {
			/* ! this is a special character meaning 
			 * I requires every service providing name, not 
			 * only the first one */
			temp++;
			prt->proreq_t = REQUIRES_EVERY_PROV;
			if(*temp == '\0') {
				/* @#! users ! */
				free(prt);
				continue;
			}
		} else {
			if (!strcmp("*",temp)) {
				prt->proreq_t = REQUIRES_ALL;
			} else {
				prt->proreq_t = PROREQ_NORMAL;
			}
		}
		strncpy(prt->name, temp, NAME_LENGTH);
		prt->next = *list;
		*list = prt;
	}
}				
#if 0
void print_service(struct service * serv) {
	struct proreq * i;
	struct pserv * j;
	printf("================\n");
	printf("        name: %s\n",serv->name);
        printf("        filename: %s\n",serv->file);
	printf("	Started: %d\n",serv->started);
	printf("	Old_init: %d\n",serv->old_init);
	for(i = serv->provides; i != NULL; i = i->next)
		printf("	Provides: %s\n",i->name);
	for(i = serv->requires; i != NULL; i = i->next) {
		if(strlen(i->name)) {
			printf("	Requires: %s (%d)\n",i->name,i->proreq_t);
			for(j = i->servl; j != NULL; j = j->next) 
			       printf("         -> %s\n",j->serv->name);
		}
	}	
}
#endif

/* fill the "list" service list with every service found in "dir" 
 * the list is sorted */
void fill_service_list(struct service ** list, char * dir) {
	char ** filelist = list_files(dir);
	char ** temp;
	for(temp = filelist; temp != NULL && *temp != NULL; temp++){
		char buffer[1024];
		int s;
		struct service * serv;
		malloc_or_die(serv,sizeof(struct service));
		s = readlink(*temp, buffer, sizeof(buffer) - 1 );
		if(s == -1 || s == 0) {
			printf("Cannot readlink %s\n",*temp);
			strncpy(buffer,*temp,sizeof(buffer));
		} else 
			buffer[s] = '\0';
		serv->requires = NULL;
		serv->provides = NULL;
		strncpy(serv->name, strrchr(buffer, '/') + 1, NAME_LENGTH);
		strncpy(serv->file, *temp, NAME_LENGTH);
		if(check_old_script(serv->file)) {
			serv->old_init = 1;
		} else {	
			serv->old_init = 0;
			parse_file_provide(serv->file,buffer,sizeof(buffer));
			fill_proreq_list(&serv->provides,buffer);
			parse_file_dep(serv->file,buffer,sizeof(buffer));
			fill_proreq_list(&serv->requires,buffer);
		}
		serv->started = 0;
		pthread_mutex_init(&serv->mut,NULL);
		pthread_cond_init(&serv->cond,NULL);
		serv->next = *list;
		*list = serv;
		free(*temp);
	}
	free(filelist);
}

/* look up who provides "name" service 
 * return a list of services pointer */
struct pserv * who_provides(char * name) {
	struct pserv * r = NULL;
	struct service * i;
	
	for(i = servlist; i != NULL; i = i->next) {
		struct proreq * j;
		if(!strcasecmp(name,i->name)) {
			struct pserv * t;
			malloc_or_die(t,sizeof(struct pserv));
			t->serv = i;
			t->next = r;
			r = t;
			continue;
		}
			
		for(j = i->provides; j != NULL; j = j->next) {
			if(!strcasecmp(name, j->name)) {
				struct pserv * t;
				malloc_or_die(t,sizeof(struct pserv));
				t->serv = i;
				t->next = r;
				r = t;
				break;
			}
		}
	}
	/* NOTFOUND */
#ifdef DEBUG
	if(r == NULL)
		printf("Requires %s NOT FOUND!\n",name);
#endif
	return r;
}
				
void clean_graph(void) {
	/* fill 0 in every node */ 
	struct service *i;
	for (i = servlist; i != NULL; i = i->next) {
		i->count = 0;
	}
}	

/* Thanks misc for the algorithm */
/* We walk on the graph recursively until
 *  - We search the end of the graph
 *  - We pass twice on a same node
 */
void run_graph(struct service * serv) {
	struct proreq * i;
	if(++serv->count == 1) { 
		for(i = serv->requires; i != NULL; i = i->next) {
			struct pserv * j;
			struct pserv * t;
			/* Requires: * service are ignored */
			if(i->name[0] == '\0' || i->proreq_t == REQUIRES_ALL || (j = who_provides(i->name)) == NULL) 
				continue;
			run_graph(j->serv);
			t = j;
			j = j->next;
			free(t);
			while(j != NULL) {
				t = j;
				j = j->next;
				if(i->proreq_t == REQUIRES_EVERY_PROV) 
					run_graph(t->serv);
				free(t);
			}
		}
	}
}	

/* Check if the requires "j" of service "serv" is valid 
 * - Doesn't loop
 * - doesn't requires on "Requires: *" service 
 * This function return a list of service pointer who fullfit the requires "j" */
struct pserv * validate_requires(struct service * serv, struct proreq * j) {
	struct pserv * target;
	struct pserv * list = NULL;

	if(j->proreq_t == REQUIRES_ALL) 
		return NULL;

	target = who_provides(j->name);
	
	while(target != NULL) {
		struct pserv * t = target;
		target = target->next;
	
		if(t->serv->requires && t->serv->requires->proreq_t == REQUIRES_ALL) {
			printf("WARNING! Requires on end service ! Ignoring ! \n");
			free(t);
			continue;
		}
		clean_graph();
		run_graph(t->serv);
		if(serv->count > 0) {
			printf("WARNING! Loop detected %s -> %s\n",serv->name, t->serv->name);
			printf(" -> Ignoring requires %s\n",j->name);
			free(t);
			continue;
		}
		/* ok. no direct loop require OK 
		 * Adding it to the list */
		t->next = list;
		list = t;
	}
		
	return list;
}

/* get the filename of the "des" unix file descriptor ( not the C one )
 * put it in "buffer" */
void get_filename(int des, char * buffer, int buflen) {
	char temp[buflen];
	int t;
	snprintf(temp,buflen,"/proc/self/fd/%d",des);
	t = readlink(temp,buffer,buflen - 1);
	if(t == -1) {
		printf("Cannot find tempfile name !\n");
		buffer[0] = '\0';
	}
	buffer[t] = '\0';
}

/* return 1 if "serv" has a requires on "me", 0 otherwise */
int requires_on_me(struct service * serv, struct service * me) {
	struct proreq * i;
	struct pserv * j;
	for(i = serv->requires; i != NULL; i = i->next) {
		for(j = i->servl; j != NULL; j = j->next) {
			if(j->serv == me)
				return 1;
		}
	}
	return 0;
}
	
/* check if the "subsys" subsystem is already up
 * return 1 if true, 0 otherwise */
int subsys_already_up(char * subsys) {
	char file[255];
	struct stat sbuf;
#ifdef DEBUG
	return mode == MODE_START ? 0 : 1;
#else
	
	snprintf(file,sizeof(file),"/var/lock/subsys/%s",subsys);
	if(!stat(file,&sbuf))
		return 1;
	snprintf(file,sizeof(file),"/var/lock/subsys/%s.init",subsys);
	if(!stat(file,&sbuf))
		return 1;
	return 0;
#endif
}
	

/* this is the function which is threaded for every service */
/* it launch the service when his requires are fullfited */
void * service_thread(void * ptr) {
	struct service * serv = (struct service *) ptr;
	struct proreq *req;
	int file;
	FILE * f;
	int already_up = 0;
	char tempfile[NAME_LENGTH];
	char command[1024];
	char tmpmask[] = "/tmp/init.XXXXXX";
	char action[6];

	if(mode == MODE_START) {
		strcpy(action,"start"); 
		for (req = serv->requires; req != NULL; req = req->next) {
			struct pserv * t;
			if(req->proreq_t == REQUIRES_ALL) {
				/* This mean this service is a "end" service
			 	* it is started after all other by the main() function */
				return NULL;
			}
		
			for(t = req->servl; t != NULL; t = t->next) {
			
				pthread_mutex_lock(&t->serv->mut);
				/* no need to loop to recheck the condition after
		 		* the condition event, the value never do a started -> not started switch */
				if(!t->serv->started) 
					pthread_cond_wait(&t->serv->cond,&t->serv->mut);
				pthread_mutex_unlock(&t->serv->mut);
				if(req->proreq_t != REQUIRES_EVERY_PROV)
					break;
			}
		}
	} else {
		/* MODE KILL */
		/* We are in stop mode, so we must wait until the end of 
		 * every service who requries us to be able to stop us */
		/* Yes, this algorith doesn't respect the % and non-% requires
		 * difference and act if every require was tagged with %.
		 * But well, this is not so important for stopping service
		 * so, who care ? */
		struct service * i;

		strcpy(action,"stop");

		if(serv->started) 
			/* yes, it CAN be possible, Require: * script are started
			 * before creating threads */
			return NULL;

		for(i = servlist; i != NULL; i = i->next) {
			if(requires_on_me(i,serv)) {
				pthread_mutex_lock(&i->mut);
				if(!i->started) 
					pthread_cond_wait(&i->cond,&i->mut);
				pthread_mutex_unlock(&i->mut);
			}
		}
	}
	
	/* All our requires are started 
	 * Let's go ! */

	already_up = subsys_already_up(serv->name);
	
	if((!already_up && mode == MODE_START) || (already_up && mode == MODE_KILL)) {
	
		strcpy(tmpmask,"/tmp/init.XXXXXX");
		if((file = mkstemp(tmpmask)) == -1) {
			printf("Cannot create temporary file !\n");
			strcpy(tempfile,"/dev/null");	
		} else {
			/* get filename from descriptor */
			get_filename(file,tempfile,sizeof(tempfile));
			if(tempfile[0] == '\0') {
				printf("Cannot get filename for temporary file !\n");
				strcpy(tempfile,"/dev/null");
			}			
			close(file);
		}
		/* run our script ! */
		if (snprintf(command,sizeof(command),"%s %s 2>&1 > %s",serv->file, action, tempfile) < sizeof(command)) {
#ifdef DEBUG
		    puts(command);
#else
		    system(command);
#endif
		} else {
		    printf("Can't run command %s: too long\n", serv->file);
		    exit(1); /* panic. not very clean */
		}
	}
	/* tell other we are done */
	pthread_mutex_lock(&serv->mut);
	serv->started = 1;
       	pthread_cond_broadcast(&serv->cond);
	pthread_mutex_unlock(&serv->mut);
	rc_splash(serv->name);
	
	if((!already_up && mode == MODE_START) || (already_up && mode == MODE_KILL)) {
		size_t n;
		if((f = fopen(tempfile,"r")) == NULL) {
			printf("Cannot open output file: %s\n",strerror(errno));
			unlink(tempfile);
			return NULL;
		}	 
	
		/* print message on the console */
		pthread_mutex_lock(&console_mutex);
		while((n = fread(command,1,sizeof(command),f)) != 0)
			fwrite(command,1,n,stdout);
		pthread_mutex_unlock(&console_mutex);
		fclose(f);
		unlink(tempfile);
	}
	return NULL;
}


/* return the number of service loaded */
/* and populate the service list ( it solve every dependency
 * so, after calling fill_requries, the system state is OK */
int fill_requires(char * directory) {
	struct service *i;
	struct proreq *j;
	int nbserv = 0;
	fill_service_list(&servlist,directory);
	
	for(i = servlist; i != NULL; i = i->next) {
		if(i->old_init == 1) {
			/* this is a non-ported init script */
			/* requires everything which is before in the list
			 * and make requires everything which is after in the list
			 * requires himself, so we keep the lsb standard compatibility */
			/* For MODE_KILL we do exactly the oposite since the requires will
			 * be inversed lated */
			/* for now, we put only requries on name, and we don't "resolve"
			 * our requries since it will be done after */
			struct service *t = servlist;
			printf("Activating Compat-mode for %s\n",i->name);
			for(;t != i; t = t->next) {
				/* do not put requires on end script */
				if(t->requires && t->requires->proreq_t == REQUIRES_ALL)
					continue;
				
				struct proreq * prt;
				malloc_or_die(prt,sizeof(struct proreq));
				
				prt->servl = NULL;
				prt->proreq_t = PROREQ_NORMAL;
				if(mode == MODE_START) {
					strcpy(prt->name,t->name);
					prt->next = i->requires;
					i->requires = prt;
				} else {
					strcpy(prt->name,i->name);
					prt->next = t->requires;
					t->requires = prt;
				}
			}
			/* put a requires on i in every service next in the list */
			for(t = t->next; t != NULL; t = t->next) {
				/* do not put requires on end script */
				if(t->requires && t->requires->proreq_t == REQUIRES_ALL)
					continue;
				
				struct proreq * prt;
				malloc_or_die(prt,sizeof(struct proreq));

				prt->servl = NULL;
				prt->proreq_t = PROREQ_NORMAL;
				if(mode == MODE_START) {
					strcpy(prt->name,i->name);
					prt->next = t->requires;
					t->requires = prt;
				} else {
					strcpy(prt->name,t->name);
					prt->next = i->requires;
					i->requires = prt;
				}
			}
		}
	}
		
	for(i = servlist; i != NULL; i = i->next) {
		nbserv++;
		for(j = i->requires; j != NULL; j = j->next) {
			if(j->proreq_t == REQUIRES_ALL) {
				struct pserv * t = j->servl;
				while(t != NULL) {
					struct pserv * temp = t;
					t = t->next;
					free(temp);
				}
				j->servl = NULL;
				continue;
			}
					
			j->servl = validate_requires(i,j);
			if(j->servl == NULL) 
				j->name[0] = '\0';
		}
	}
	return nbserv;
}	

/* start the system */
void mode_start(char * directory) {
	struct service *i;
	pthread_t * thread_tab;
	int nbserv = 0;
	int p;

	mode = MODE_START;

	nbserv = fill_requires(directory);
	
	malloc_or_die(thread_tab, nbserv * sizeof(pthread_t));
	
	for(i = servlist, nbserv = 0; i != NULL; i = i->next, nbserv++) {
		pthread_create(&thread_tab[nbserv], NULL,service_thread,(void *) i);
	}
	/* wait for each service end */
	for(p = 0;p < nbserv; p++)
		pthread_join(thread_tab[p],NULL);
	free(thread_tab);
	
	/* Start services who are still not started ( thoses with Requires: * ) */
	for(i = servlist; i != NULL; i = i->next) {
		if(!i->started) {
			rc_splash(i->name);
			if(!subsys_already_up(i->name)) {
				char command[1024];
				if (snprintf(command,sizeof(command),"%s start 2>&1",i->file) < sizeof(command)) {
#ifdef DEBUG
				    puts(command);
#else
				    system(command);
#endif
				} else {
				    printf("Can't run command %s: too long\n", i->file);
				    exit(1); /* panic. not very clean */
				}
			}
		}
	}
}

/* stop de system */
void mode_kill(char * directory) {
	struct service *i;
	pthread_t * thread_tab;
	int nbserv = 0;
	int p;
	
	mode = MODE_KILL;

	nbserv = fill_requires(directory);

	/* run every Require: * service first */

	for(i = servlist; i != NULL; i = i->next) {
		if(i->requires && i->requires->proreq_t == REQUIRES_ALL) {
			char command[1024];
			if (snprintf(command,sizeof(command),"%s stop 2>&1",i->file) < sizeof(command)) {
				if(subsys_already_up(i->name)) {
#ifdef DEBUG
					puts(command);
#else
					system(command);
#endif
				}
				i->started = 1;
				rc_splash(i->name);
			} else {
				printf("Can't run command %s: too long\n", i->file);
				exit(1);
			}
		}
	}

	malloc_or_die(thread_tab, nbserv * sizeof(pthread_t));
	
	for(i = servlist, nbserv = 0; i != NULL; i = i->next, nbserv++) {
		pthread_create(&thread_tab[nbserv], NULL,service_thread,(void *) i);
	}
	/* wait for each service end */
	for(p = 0;p < nbserv; p++)
		pthread_join(thread_tab[p],NULL);
	free(thread_tab);
}

int main(int argc, char ** argv) {

	char * env;
	
	/* first argument is the directory where the script are located */
	if(argc != 3) {
		printf("Usage: %s [mode] [directory]\n",argv[0]);
		exit(1);
	}

	if(*argv[1] == MODE_START) 
		mode_start(argv[2]);
	else if (*argv[1] == MODE_KILL)
		mode_kill(argv[2]);
	else {
		printf("Unknow mode %s\n",argv[1]);
		exit(1);
	}

	if((env = getenv("progress")) != NULL) 
		return atoi(env);
	
	return 0;
}


